import logging

from datapackage_pipelines.status import status
from .celery_app import celery_app
from ..specs import pipelines, PipelineSpec
from ..manager.tasks import execute_pipeline

executed_hashes = {}


@celery_app.task
def update_pipelines(action, completed_pipeline_id, completed_trigger):
    # action=init: register all pipelines, trigger anything that's dirty
    # action=update: iterate over all pipelines, register new ones, trigger dirty ones
    # action=complete: iterate over all pipelines, trigger dependencies
    # completed_pipeline_id: pipeline id that had just completed (when applicable)
    # completed_trigger: the trigger for the pipeline that had just completed (when applicable)
    logging.debug("Update Pipelines (%s)", action)
    status_all_pipeline_ids = set(sts['id'] for sts in status.all_statuses())

    all_pipeline_ids = set()
    for spec in pipelines():
        pipeline_id = spec.pipeline_id
        all_pipeline_ids.add(pipeline_id)

        run = False
        if action == 'init':
            status.register(spec.pipeline_id,
                            spec.cache_hash,
                            spec.pipeline_details,
                            spec.source_details,
                            spec.errors)
            if spec.dirty:
                run = True
        elif action == 'update':
            registered = True
            if spec.pipeline_id not in status_all_pipeline_ids:
                registered = status.register(spec.pipeline_id,
                                             spec.cache_hash,
                                             spec.pipeline_details,
                                             spec.source_details,
                                             spec.errors)
                logging.info("NEW Pipeline: %s (registered? %s)", spec, registered)
            logging.debug('Pipeline: %s (dirty: %s, %s != %s?)',
                          spec.pipeline_id, spec.dirty, executed_hashes.get(spec.pipeline_id), spec.cache_hash)
            if registered and spec.dirty and executed_hashes.get(spec.pipeline_id) != spec.cache_hash:
                executed_hashes[spec.pipeline_id] = spec.cache_hash
                run = True
        elif action == 'complete':
            if completed_pipeline_id in spec.dependencies:
                if spec.dirty or completed_trigger == 'schedule':
                    status.register(spec.pipeline_id,
                                    spec.cache_hash,
                                    spec.pipeline_details,
                                    spec.source_details,
                                    spec.errors)
                    run = True

        if len(spec.errors) == 0 and run:
            logging.info('Executing task %s (from action "%s")', pipeline_id, action)
            pipeline_status = status.queued(pipeline_id)
            execute_pipeline_task.delay(pipeline_id,
                                        spec.pipeline_details,
                                        spec.path,
                                        'dirty-task' if completed_trigger is None else completed_trigger,
                                        pipeline_status.data['queued'])

    extra_pipelines = status_all_pipeline_ids.difference(all_pipeline_ids)
    for pipeline_id in extra_pipelines:
        logging.info("Removing Pipeline: %s", pipeline_id)
        status.deregister(pipeline_id)


@celery_app.task
def execute_scheduled_pipeline(pipeline_id):
    for spec in pipelines():
        if spec.pipeline_id == pipeline_id:
            if len(spec.errors) == 0:
                status.register(spec.pipeline_id,
                                spec.cache_hash,
                                spec.pipeline_details,
                                spec.source_details,
                                spec.errors)
                logging.info('Executing SCHEDULED task %s', pipeline_id)
                execute_pipeline_task.delay(pipeline_id,
                                            spec.pipeline_details,
                                            spec.path,
                                            'scheduled',
                                            0)
            else:
                logging.warning('Skipping SCHEDULED task %s, as it has errors %r', pipeline_id, spec.errors)
            break


@celery_app.task
def execute_pipeline_task(pipeline_id,
                          pipeline_details,
                          pipeline_cwd,
                          trigger,
                          queue_time):

    last_queue_time = status.get_status(pipeline_id).get('queued')
    if queue_time in {0, last_queue_time}:
        spec = PipelineSpec(pipeline_id=pipeline_id,
                            pipeline_details=pipeline_details,
                            path=pipeline_cwd)
        success, _ = \
            execute_pipeline(spec,
                             trigger,
                             False)

        if success:
            update_pipelines.delay('complete', pipeline_id, trigger)



import logging
import os
import time

from .backend_redis import RedisBackend
from .backend_sqlite import SqliteBackend


class PipelineStatus(object):

    STATES = {
        'INIT': {},
        'REGISTERED': {
            'success': None,
            'message': "Didn't run",
            'reason': "Didn't run yet"
        },
        'INVALID': {
            'success': False,
        },
        'RUNNING': {
            'success': None,
            'message': 'Running',
        },
        'SUCCEEDED': {
            'success': True,
            'message': 'Succeeded',
        },
        'FAILED': {
            'success': False,
            'message': 'Failed'
        }
    }

    def __init__(self, backend, pipeline_id):
        self.backend = backend
        self.pipeline_id = pipeline_id
        self.data = backend.get_status(pipeline_id)
        if self.data is None or 'state' not in self.data:
            self.data = {'state': 'INIT'}
        self.data.update({
            'id': self.pipeline_id,
        })

    def set_state(self, state):
        self.data['state'] = state
        self.data['dirty'] = False
        self.data.update(self.STATES[state])

    def check_running(self):
        cur_time = time.time()
        if self.data is None:
            return False

        # A running task must have been updated in the last 10 seconds
        running = self.data.get('state') == 'RUNNING'
        updating = (cur_time - self.data.get('updated', 0)) < 60
        if running and not updating:
            self.set_idle(False, force=True)
        return self.data.get('state') == 'RUNNING'

    def check_successful(self):
        return self.data.get('state') == 'SUCCEEDED'

    def check_waiting(self):
        return self.data.get('state') in {'REGISTERED', 'INVALID'}

    def set_running(self, trigger, log):
        if self.data['state'] not in {'REGISTERED',
                                      'SUCCEEDED',
                                      'FAILED',
                                      'RUNNING'}:
            logging.error('set_running: bad state %s', self.data['state'])
            return

        cur_time = time.time()
        if not self.check_running():
            self.data.update({
                'started': cur_time,
                'trigger': trigger,
            })
            self.set_state('RUNNING')

        self.data.update({
            'reason': log
        })
        self.save()

    def set_idle(self, success,
                 log=None, cache_hash=None, stats=None, force=False):
        if self.data['state'] not in {'RUNNING'}:
            logging.error('set_idle: bad state %s', self.data['state'])
            return

        if stats is None:
            stats = {}

        cur_time = time.time()
        if force or self.check_running():
            self.data.update({
                'ended': cur_time,
                'reason': log,
            })
            if success is True:
                self.data.update({
                    'last_success': self.data['ended'],
                    'cache_hash': cache_hash,
                    'stats': stats,
                })
                self.set_state('SUCCEEDED')
            else:
                self.set_state('FAILED')
        self.save()

    def set_queued(self):
        cur_time = time.time()
        self.data.update({
            'queued': cur_time,
        })
        self.save()

    def register(self, cache_hash, pipeline, source, errors):

        if self.check_running():
            return False

        # Is pipeline dirty?
        dirty = self.data.setdefault('cache_hash', '') != cache_hash
        dirty = dirty or self.data.get('state') != 'SUCCEEDED'
        dirty = dirty and len(errors) == 0

        self.data.update({
            'pipeline': pipeline,
            'source': source,
            'dirty': False
        })
        if len(errors) > 0:
            self.data.update({
                'message': errors[0][0],
                'reason': '\n'.join('{}: {}'.format(*e) for e in errors),
            })
            self.set_state('INVALID')
        else:
            if self.data.get('state') in {'INIT', 'INVALID'}:
                self.set_state('REGISTERED')
            self.data['dirty'] = dirty

        self.backend.register_pipeline_id(self.pipeline_id)
        self.save()
        return dirty

    def deregister(self):
        self.backend.deregister_pipeline_id(self.pipeline_id)

    def save(self):
        cur_time = time.time()
        self.data.update({
            'updated': cur_time
        })
        self.backend.set_status(self.pipeline_id, self.data)


class StatusManager(object):

    def __init__(self, host=None, port=6379):
        self._host = host
        self._port = port
        self._backend = None

    @property
    def backend(self):
        if self._backend is None:
            redis = RedisBackend(self._host, self._port)
            self._backend = redis if redis.is_init() else SqliteBackend()
        return self._backend

    def is_running(self, _id):
        return PipelineStatus(self.backend, _id).check_running()

    def is_waiting(self, _id):
        return PipelineStatus(self.backend, _id).check_waiting()

    def is_successful(self, _id):
        return PipelineStatus(self.backend, _id).check_successful()

    def running(self, _id, trigger=None, log=None):
        PipelineStatus(self.backend, _id).set_running(trigger, log)

    def idle(self, _id, success, reason, cache_hash, stats):
        PipelineStatus(self.backend, _id) \
            .set_idle(success, reason, cache_hash, stats)

    def register(self, _id, cache_hash, pipeline=(), source=None, errors=()):
        return PipelineStatus(self.backend, _id) \
            .register(cache_hash, pipeline, source, errors)

    def deregister(self, _id):
        return PipelineStatus(self.backend, _id) \
            .deregister()

    def queued(self, _id):
        ps = PipelineStatus(self.backend, _id)
        ps.set_queued()
        return ps

    def initialize(self):
        self.backend.reset()

    def get_status(self, _id):
        return self.backend.get_status(_id)

    def all_statuses(self):
        return self.backend.all_statuses()


status = StatusManager(os.environ.get('DPP_REDIS_HOST'))



import os
import unittest

import datapackage_pipelines.lib
from datapackage_pipelines.utilities.lib_test_helpers \
    import mock_processor_test

import logging
log = logging.getLogger(__name__)


class TestStandardProcessors(unittest.TestCase):

    def test_add_metadata_processor(self):
        # Input arguments used by our mock `ingest`
        datapackage = {
            'name': 'my-datapackage',
            'project': 'my-project',
            'resources': []
        }
        params = {
            'hello': 'world'
        }

        # Path to the processor we want to test
        processor_dir = \
            os.path.dirname(datapackage_pipelines.lib.__file__)
        processor_path = os.path.join(processor_dir, 'add_metadata.py')

        # Trigger the processor with our mock `ingest` and capture what it will
        # returned to `spew`.
        spew_args, _ = \
            mock_processor_test(processor_path,
                                (params, datapackage, []))

        # Get the returned datapackage and res_iter after calling our processor
        spew_dp = spew_args[0]
        spew_res_iter = spew_args[1]

        assert spew_dp == {'name': 'my-datapackage', 'project': 'my-project',
                           'resources': [], 'hello': 'world'}
        assert spew_res_iter == []



import sys
import copy
import logging

import datapackage
from jsontableschema.exceptions import InvalidCastError
from jsontableschema.model import SchemaModel

from ..utilities.extended_json import json


class ResourceIterator(object):

    def __init__(self, infile, spec, orig_spec,
                 validate=False, debug=False):
        self.spec = spec
        self.table_schema = SchemaModel(orig_spec['schema'])
        self.validate = validate
        self.infile = infile
        self.debug = debug
        self.stopped = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.stopped:
            raise StopIteration()
        if self.debug:
            logging.error('WAITING')
        line = self.infile.readline().strip()
        if self.debug:
            logging.error('INGESTING: %r', line)
        if line == '':
            self.stopped = True
            raise StopIteration()
        line = json.loadl(line)
        if self.validate:
            for k, v in line.items():
                try:
                    self.table_schema.cast(k, v)
                except (InvalidCastError, TypeError):
                    field = self.table_schema.get_field(k)
                    if field is None:
                        raise ValueError('Validation failed: No such field %s',
                                         k)
                    else:
                        raise ValueError('Validation failed: Bad value %r '
                                         'for field %s with type %s',
                                         v, k, field.get('type'))

        return line

    def next(self):
        return self.__next__()


def process_input(infile, validate=False, debug=False):

    dp_json = infile.readline().strip()
    if dp_json == '':
        sys.exit(-3)
    dp = json.loads(dp_json)
    resources = dp.get('resources', [])
    original_resources = copy.deepcopy(resources)

    profiles = list(dp.get('profiles', {}).keys())
    profile = 'tabular'
    if 'tabular' in profiles:
        profiles.remove('tabular')
    if len(profiles) > 0:
        profile = profiles.pop(0)
    schema = datapackage.schema.Schema(profile)
    schema.validate(dp)

    infile.readline().strip()

    def resources_iterator(_resources, _original_resources):
        # we pass a resource instance that may be changed by the processing
        # code, so we must keep a copy of the original resource (used to
        # validate incoming data)
        ret = []
        for resource, orig_resource in zip(_resources, _original_resources):
            if 'path' not in resource:
                continue

            res_iter = ResourceIterator(infile,
                                        resource, orig_resource,
                                        validate, debug)
            ret.append(res_iter)
        return iter(ret)

    return dp, resources_iterator(resources, original_resources)



import os
import logging
from jsontableschema_sql import Storage
from sqlalchemy import create_engine

from datapackage_pipelines.lib.dump.dumper_base import DumperBase


class SQLDumper(DumperBase):

    def initialize(self, parameters):
        super(SQLDumper, self).initialize(parameters)
        table_to_resource = parameters['tables']
        engine = parameters.get('engine', 'env://DPP_DB_ENGINE')
        if engine.startswith('env://'):
            env_var = engine[6:]
            engine = os.environ.get(env_var)
            assert engine is not None, \
                "Couldn't connect to DB - " \
                "Please set your '%s' environment variable" % env_var
        self.engine = create_engine(engine)

        for k, v in table_to_resource.items():
            v['table-name'] = k

        self.converted_resources = \
            dict((v['resource-name'], v) for v in table_to_resource.values())

    def handle_resource(self, resource, spec, parameters, datapackage):
        resource_name = spec['name']
        if resource_name not in self.converted_resources:
            return resource
        else:
            converted_resource = self.converted_resources[resource_name]
            mode = converted_resource.get('mode', 'rewrite')
            table_name = converted_resource['table-name']
            storage = Storage(self.engine, prefix=table_name)
            if mode == 'rewrite' and '' in storage.buckets:
                storage.delete('')
            if '' not in storage.buckets:
                logging.info('Creating DB table %s', table_name)
                storage.create('', spec['schema'])
            update_keys = None
            if mode == 'update':
                update_keys = converted_resource.get('update_keys')
                if update_keys is None:
                    update_keys = spec['schema'].get('primaryKey', [])
            logging.info('Writing to DB %s -> %s (mode=%s, keys=%s)',
                         resource_name, table_name, mode, update_keys)
            return storage.write('', resource,
                                 keyed=True, as_generator=True,
                                 update_keys=update_keys)


SQLDumper()()



import os

from datapackage_pipelines.status import status

from .resolver import resolve_executor
from .errors import SpecError
from .schemas.validator import validate_pipeline
from .parsers import BasicPipelineParser, SourceSpecPipelineParser
from .hashers import HashCalculator, DependencyMissingException

SPEC_PARSERS = [
    BasicPipelineParser(),
    SourceSpecPipelineParser()
]


def resolve_processors(spec):
    abspath = os.path.abspath(spec.path)
    for step in spec.pipeline_details['pipeline']:
        if 'executor' not in step:
            step['executor'] = resolve_executor(step,
                                                abspath,
                                                spec.errors)


def process_schedules(spec):
    if spec.schedule is None:
        schedule = spec.pipeline_details.get('schedule', {})
        if 'crontab' in schedule:
            schedule = schedule['crontab'].split()
            spec.schedule = schedule


def calculate_dirty(spec):
    pipeline_status = status.get_status(spec.pipeline_id) or {}
    dirty = pipeline_status.get('cache_hash', '') != spec.cache_hash
    dirty = dirty or pipeline_status.get('state') != 'SUCCEEDED'
    dirty = dirty and len(spec.errors) == 0

    spec.dirty = dirty


def find_specs(root_dir='.'):
    for dirpath, _, filenames in os.walk(root_dir):
        if dirpath.startswith('./.'):
            continue
        for filename in filenames:
            for parser in SPEC_PARSERS:
                if parser.check_filename(filename):
                    yield from parser.to_pipeline(os.path.join(dirpath, filename))


def pipelines():

    specs = find_specs()
    hasher = HashCalculator()
    deferred_amount = 0
    while specs is not None:
        deferred = []

        for spec in specs:
            if (spec.pipeline_details is not None and
                    validate_pipeline(spec.pipeline_details, spec.errors)):

                resolve_processors(spec)
                process_schedules(spec)

                if len(spec.errors) == 0:
                    try:
                        hasher.calculate_hash(spec)
                    except DependencyMissingException:
                        deferred.append(spec)
                        continue

                calculate_dirty(spec)

            yield spec

        if len(deferred) > 0:
            if len(deferred) == deferred_amount:
                for spec in deferred:
                    spec.errors.append(
                        SpecError('Missing dependency',
                                  'Failed to find a pipeline dependency')
                    )
            deferred_amount = len(deferred)
            specs = iter(deferred)
        else:
            specs = None


def register_all_pipelines():
    for spec in pipelines():
        status.register(spec.pipeline_id,
                        spec.cache_hash,
                        pipeline=spec.pipeline_details,
                        source=spec.source_details,
                        errors=spec.errors)



import itertools

import datapackage

from datapackage_pipelines.wrapper import ingest, spew

parameters, dp, res_iter = ingest()

url = parameters['url']
resource = parameters['resource']

selected_resource = None
datapackage = datapackage.DataPackage(url)
for i, orig_res in enumerate(datapackage.resources):
    if resource == i or resource == orig_res.descriptor.get('name'):
        dp['resources'].append(orig_res.descriptor)
        selected_resource = orig_res
        break

assert selected_resource is not None, "Failed to find resource with index or name matching %r" % resource

spew(dp, itertools.chain(res_iter, [selected_resource.iter()]))



import datetime

import slugify
import yaml
import mistune

from flask import Flask, render_template, abort, redirect
from flask_cors import CORS
from flask_jsonpify import jsonify

from datapackage_pipelines.status import status
from datapackage_pipelines.specs import register_all_pipelines

register_all_pipelines()

app = Flask(__name__)
app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True
CORS(app)


def datestr(x):
    return str(datetime.datetime.fromtimestamp(x))


def yamlize(x):
    ret = yaml.dump(x, default_flow_style=False)
    return ret


markdown = mistune.Markdown(hard_wrap=True)


def make_hierarchies(statuses):

    def group(l):
        pipelines = list(filter(lambda x: len(x['id']) == 1, l))
        children_ = filter(lambda x: len(x['id']) > 1, l)
        groups_ = {}
        for child in children_:
            child_key = child['id'].pop(0)
            groups_.setdefault(child_key, []).append(child)
        children_ = dict(
            (k, group(v))
            for k, v in groups_.items()
        )
        for p in pipelines:
            p['id'] = p['id'][0]
        return {
            'pipelines': pipelines,
            'children': children_
        }

    def flatten(children_):
        for k, v in children_.items():
            v['children'] = flatten(v['children'])
            child_keys = list(v['children'].keys())
            if len(child_keys) == 1:
                child_key = child_keys[0]
                children_['/'.join([k, child_key])] = v['children'][child_key]
                del children_[k]
        return children_

    statuses = sorted(statuses, key=lambda x: x['id'])
    statuses = [
        {
            'id': st['id'].split('/'),
            'title': st.get('pipeline', {}).get('title'),
            'stats': st.get('stats'),
            'slug': st.get('slug')
        }
        for st in statuses
    ]
    groups = group(statuses)
    children = groups.get('children', {})
    groups['children'] = flatten(children)

    return groups


@app.route("/")
def main():
    statuses = sorted(status.all_statuses(), key=lambda x: x.get('id'))
    for pipeline in statuses:
        for key in ['ended', 'last_success', 'started']:
            if pipeline.get(key):
                pipeline[key] = datestr(pipeline[key])
        pipeline['class'] = {'INIT': 'primary',
                             'REGISTERED': 'primary',
                             'INVALID': 'danger',
                             'RUNNING': 'warning',
                             'SUCCEEDED': 'success',
                             'FAILED': 'danger'
                             }[pipeline.get('state', 'INIT')]

        pipeline['slug'] = slugify.slugify(pipeline['id'])
        pipeline['id'] = pipeline['id'].lstrip('./')

    def state_and_not_dirty(state, p):
        return p.get('state') == state and not p.get('dirty')

    def state_or_dirty(state, p):
        return p.get('state') == state or p.get('dirty')

    categories = [
        ['ALL', 'All Pipelines', lambda _, __: True],
        ['INVALID', "Can't start", state_and_not_dirty],
        ['REGISTERED', 'Waiting to run', state_or_dirty],
        ['RUNNING', 'Running', state_and_not_dirty],
        ['FAILED', 'Failed Execution', state_and_not_dirty],
        ['SUCCEEDED', 'Successful Execution', state_and_not_dirty],
    ]
    for item in categories:
        item.append([p for p in statuses
                     if item[2](item[0], p)])
        item.append(len(item[-1]))
        item.append(make_hierarchies(item[-2]))
    return render_template('dashboard.html',
                           categories=categories,
                           yamlize=yamlize,
                           markdown=markdown)


@app.route("/api/<field>/<path:pipeline_id>")
def pipeline_api(field, pipeline_id):
    fields = {
        'log': 'reason',
        'pipeline': 'pipeline',
        'source': 'source'
    }
    field = fields.get(field)
    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get_status(pipeline_id)
    if pipeline_status is None or field is None:
        abort(404)
    ret = pipeline_status[field]
    if field != 'reason':
        ret = yamlize(ret)
    ret = ret.split('\n')
    ret = {'text': ret}
    return jsonify(ret)


@app.route("/badge/<path:pipeline_id>")
def badge(pipeline_id):
    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get_status(pipeline_id)
    if pipeline_status is None:
        abort(404)
    status_text = pipeline_status.get('message')
    success = pipeline_status.get('success')
    if success is True:
        record_count = pipeline_status.get('stats', {}).get('total_row_count')
        if record_count is not None:
            status_text += ' (%d records)' % record_count
        status_color = 'brightgreen'
    elif success is False:
        status_color = 'red'
    else:
        status_color = 'lightgray'
    return redirect('https://img.shields.io/badge/{}-{}-{}.svg'.format(
        'pipeline', status_text, status_color
    ))



import os
import tempfile
import collections
import cachetools
import logging

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.extended_json import json

try:
    import plyvell as DB_ENGINE
    logging.info('Using leveldb for joining')
    db_kind = 'LevelDB'
except ImportError:
    import sqlite3 as DB_ENGINE
    logging.info('Using sqlite for joining')
    db_kind = 'sqlite'


class KeyCalc(object):

    def __init__(self, key_spec):
        if isinstance(key_spec, list):
            key_spec = ':'.join('{%s}' % key for key in key_spec)
        self.key_spec = key_spec

    def __call__(self, row):
        return self.key_spec.format(**row)


class SqliteDB(object):

    def __init__(self):
        self.tmpfile = tempfile.NamedTemporaryFile()
        self.db = DB_ENGINE.connect(self.tmpfile.name)
        self.cursor = self.db.cursor()
        self.cursor.execute('''CREATE TABLE d (key text, value text)''')
        self.cursor.execute('''CREATE UNIQUE INDEX i ON d (key)''')

    def get(self, key):
        ret = self.cursor.execute('''SELECT value FROM d WHERE key=?''',
                                  (key,)).fetchone()
        if ret is None:
            raise KeyError()
        else:
            return json.loads(ret[0])

    def set(self, key, value):
        value = json.dumps(value)
        try:
            self.get(key)
            self.cursor.execute('''UPDATE d SET value=? WHERE key=?''',
                                (value, key))
        except KeyError:
            self.cursor.execute('''INSERT INTO d VALUES (?, ?)''',
                                (key, value))
        self.db.commit()

    def keys(self):
        cursor = self.db.cursor()
        keys = cursor.execute('''SELECT key FROM d ORDER BY key ASC''')
        for key, in keys:
            yield key


class LevelDB(object):

    def __init__(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.db = DB_ENGINE.DB(self.tmpdir.name, create_if_missing=True)

    def get(self, key):
        ret = self.db.get(key.encode('utf8'))
        if ret is None:
            raise KeyError()
        else:
            return json.loads(ret.decode('utf8'))

    def set(self, key, value):
        value = json.dumps(value).encode('utf8')
        key = key.encode('utf8')
        self.db.put(key, value)

    def keys(self):
        for key, value in self.db:
            yield key.decode('utf8')


DB = LevelDB if db_kind == 'LevelDB' else SqliteDB


class CachedDB(cachetools.LRUCache):

    def __init__(self):
        super(CachedDB, self).__init__(1024, self._dbget)
        self.db = DB()

    def popitem(self):
        key, value = super(CachedDB, self).popitem()
        self._dbset(key, value)
        return key, value

    def _dbget(self, key):
        value = self.db.get(key)
        return value

    def _dbset(self, key, value):
        assert value is not None
        self.db.set(key, value)

    def sync(self):
        for key in iter(self):
            value = cachetools.Cache.__getitem__(self, key)
            self._dbset(key, value)

    def keys(self):
        self.sync()
        return self.db.keys()


db = CachedDB()


def identity(x):
    return x


Aggregator = collections.namedtuple('Aggregator',
                                    ['func', 'finaliser', 'dataType'])
AGGREGATORS = {
    'sum': Aggregator(lambda curr, new:
                      new + curr if curr is not None else new,
                      identity,
                      None),
    'avg': Aggregator(lambda curr, new:
                      (curr[0] + 1, new + curr[1])
                      if curr is not None
                      else (1, new),
                      lambda value: value[1] / value[0],
                      None),
    'max': Aggregator(lambda curr, new:
                      max(new, curr) if curr is not None else new,
                      identity,
                      None),
    'min': Aggregator(lambda curr, new:
                      min(new, curr) if curr is not None else new,
                      identity,
                      None),
    'first': Aggregator(lambda curr, new:
                        curr if curr is not None else new,
                        identity,
                        None),
    'last': Aggregator(lambda curr, new: new,
                       identity,
                       None),
    'count': Aggregator(lambda curr, new:
                        curr+1 if curr is not None else 1,
                        identity,
                        'integer'),
    'any': Aggregator(lambda curr, new: new,
                      identity,
                      None),
    'set': Aggregator(lambda curr, new:
                      curr.union({new}) if curr is not None else {new},
                      lambda value: list(value) if value is not None else [],
                      'array'),
    'array': Aggregator(lambda curr, new:
                        curr + [new] if curr is not None else [new],
                        identity,
                        'array'),
}

parameters, datapackage, resource_iterator = ingest()

source = parameters['source']
source_name = source['name']
source_key = KeyCalc(source['key'])
source_delete = source.get('delete', False)

target = parameters['target']
target_name = target['name']
if target['key'] is not None:
    target_key = KeyCalc(target['key'])
else:
    target_key = None
deduplication = target_key is None

fields = parameters['fields']
full = parameters.get('full', True)


def indexer(resource):
    for row in resource:
        key = source_key(row)
        try:
            current = db[key]
        except KeyError:
            current = {}
        for field, spec in fields.items():
            name = spec['name']
            curr = current.get(field)
            agg = spec['aggregate']
            if agg != 'count':
                new = row.get(name)
            else:
                new = ''
            if new is not None:
                current[field] = AGGREGATORS[agg].func(curr, new)
        db[key] = current
        yield row


def process_target(resource):
    empty_extra = dict((f, None) for f in fields.keys())
    if deduplication:
        # just empty the iterable
        collections.deque(indexer(resource), maxlen=0)
        for key in db.keys():
            row = dict(
                (f, None) for f in fields.keys()
            )
            row.update(dict(
                (k, AGGREGATORS[fields[k]['aggregate']].finaliser(v))
                for k, v in db.db.get(key).items()
            ))
            yield row
    else:
        for row in resource:
            key = target_key(row)
            try:
                extra = db[key]
            except KeyError:
                if not full:
                    continue
                extra = empty_extra
            extra = dict(
                (k, AGGREGATORS[fields[k]['aggregate']].finaliser(v))
                for k, v in extra.items()
            )
            row.update(extra)
            yield row


def new_resource_iterator(resource_iterator_):
    has_index = False
    for resource in resource_iterator_:
        name = resource.spec['name']
        if name == source_name:
            has_index = True
            if source_delete:
                # just empty the iterable
                collections.deque(indexer(resource), maxlen=0)
            else:
                yield indexer(resource)
            if deduplication:
                yield process_target(resource)
        elif name == target_name:
            assert has_index
            yield process_target(resource)
        else:
            yield resource


def process_target_resource(source_spec, resource):
    target_fields = \
        resource.setdefault('schema', {}).setdefault('fields', [])
    added_fields = sorted(fields.keys())
    for field in added_fields:
        spec = fields[field]
        if spec is None:
            fields[field] = spec = {}
        if 'name' not in spec:
            spec['name'] = field
        if 'aggregate' not in spec:
            spec['aggregate'] = 'any'
        agg = spec['aggregate']
        data_type = AGGREGATORS[agg].dataType
        if data_type is None:
            source_field = \
                next(filter(lambda f, spec_=spec:
                            f['name'] == spec_['name'],
                            source_spec['schema']['fields']))
            data_type = source_field['type']
        target_fields.append({
            'name': field,
            'type': data_type
        })
    return resource


def process_datapackage(datapackage_):

    new_resources = []
    source_spec = None

    for resource in datapackage_['resources']:

        if resource['name'] == source_name:
            source_spec = resource
            if not source_delete:
                new_resources.append(resource)
            if deduplication:
                resource = process_target_resource(
                    source_spec,
                    {
                        'name': target_name,
                        'path': os.path.join('data', target_name + '.csv')
                    })
                new_resources.append(resource)

        elif resource['name'] == target_name:
            assert isinstance(source_spec, dict), \
                "Source resource must appear before target resource"
            resource = process_target_resource(source_spec, resource)
            new_resources.append(resource)

        else:
            new_resources.append(resource)

    datapackage_['resources'] = new_resources
    return datapackage_


spew(process_datapackage(datapackage),
     new_resource_iterator(resource_iterator))
